@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 2008-2011, University of Amsterdam
+    Copyright (C): 2008-2013, University of Amsterdam
 			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
@@ -54,7 +54,7 @@ two additional modules that acts as plugins:
 
     * library(http/http_header)
     Loading this library causes http_open/3 to support the =POST= method
-    in addition to =GET= and =HEAD=.
+    in addition to =GET=, =HEAD= and =DELETE=.
 
     * library(http/http_ssl_plugin)
     Loading this library causes http_open/3 HTTPS connections.  Relevant
@@ -106,7 +106,7 @@ resource. See also parse_time/2.
 		     [ authorization(compound),
 		       final_url(-atom),
 		       header(+atom, -atom),
-		       method(oneof([get,head,post])),
+		       method(oneof([delete,get,head,post])),
 		       size(-integer),
 		       status_code(-integer),
 		       timeout(number),
@@ -124,7 +124,7 @@ resource. See also parse_time/2.
 %	Default value for =|User-Agent|=,  can   be  overruled using the
 %	option user_agent(Agent) of http_open/3.
 
-user_agent('SWI-Prolog <http://www.swi-prolog.org>').
+user_agent('SWI-Prolog').
 
 %%	http_open(+URL, -Stream, +Options) is det.
 %
@@ -155,7 +155,8 @@ user_agent('SWI-Prolog <http://www.swi-prolog.org>').
 %	  AtomValue is unified to the empty atom ('').
 %
 %	  * method(+Method)
-%	  One of =get= (default) or =head=.   The  =head= message can be
+%	  One of =get= (default), =head= or =delete=.
+%	  The  =head= message can be
 %	  used in combination with  the   header(Name,  Value) option to
 %	  access information on the resource   without actually fetching
 %	  the resource itself.  The  returned   stream  must  be  closed
@@ -183,9 +184,12 @@ user_agent('SWI-Prolog <http://www.swi-prolog.org>').
 %	  Provided if library(http/http_header) is also loaded.  Data is
 %	  handed to http_post_data/3.
 %
-%	  * proxy(+Host, +Port)
+%	  * proxy(+Host:Port)
 %	  Use an HTTP proxy to connect to the outside world.
 %
+%	  * proxy(+Host, +Port)
+%	  Synonym for proxy(+Host:Port).  Deprecated.
+%
 %	  * proxy_authorization(+Authorization)
 %	  Send authorization to the proxy.  Otherwise   the  same as the
 %	  =authorization= option.
@@ -227,7 +231,7 @@ user_agent('SWI-Prolog <http://www.swi-prolog.org>').
 
 http_open(URL, Stream, QOptions) :-
 	meta_options(is_meta, QOptions, Options),
-	(   atom(URL)
+	(   atomic(URL)
 	->  parse_url_ex(URL, Parts)
 	;   Parts = URL
 	),
@@ -242,7 +246,10 @@ is_meta(pem_password_hook).		% SSL plugin callbacks
 is_meta(cert_verify_hook).
 
 http_open_parts(Parts, Stream, Options0) :-
-	memberchk(proxy(Host, ProxyPort), Options0), !,
+	(   is_list(Options0),
+	    memberchk(proxy(Host, ProxyPort), Options0)
+	;   option(proxy(Host:ProxyPort), Options0)
+	), !,
 	parts_uri(Parts, RequestURI),
 	Options = [visited(Parts)|Options0],
 	open_socket(Host:ProxyPort, In, Out, Options),
@@ -253,7 +260,7 @@ http_open_parts(Parts, Stream, Options0) :-
 	send_rec_header(Out, In, Stream, HostPort, RequestURI, Parts, Options),
 	return_final_url(Options).
 http_open_parts(Parts, Stream, Options0) :-
-	memberchk(host(Host), Parts),
+	option(host(Host), Parts),
 	parts_scheme(Parts, Scheme),
 	default_port(Scheme, DefPort),
 	url_part(port(Port), Parts, DefPort),
@@ -348,6 +355,7 @@ method(Options, MNAME) :-
 	;   domain_error(method, M)
 	).
 
+map_method(delete, 'DELETE').
 map_method(get,  'GET').
 map_method(head, 'HEAD').
 map_method(post, 'POST') :-
@@ -377,7 +385,7 @@ x_header(_, _).
 auth_header(basic(User, Password), Header, Out) :- !,
 	format(codes(Codes), '~w:~w', [User, Password]),
 	phrase(base64(Codes), Base64Codes),
-	format(Out, '~w: basic ~s\r\n', [Header, Base64Codes]).
+	format(Out, '~w: Basic ~s\r\n', [Header, Base64Codes]).
 auth_header(Auth, _, _) :-
 	domain_error(authorization, Auth).
 
@@ -472,6 +480,7 @@ redirect_options(Options0, Options) :-
 	;   Options = Options1
 	).
 
+redirect_method(delete).
 redirect_method(get).
 redirect_method(head).
 
@@ -515,22 +524,21 @@ open_socket(Address, In, Out, Options) :-
 	      )),
 	debug(http(open), '\tok ~p --> ~p', [In, Out]),
 	set_stream(In, record_position(false)),
-	(   memberchk(timeout(Timeout), Options)
+	(   option(timeout(Timeout), Options)
 	->  set_stream(In, timeout(Timeout))
 	;   true
 	).
 
 
 return_size(Options, Lines) :-
-	memberchk(size(Size), Options), !,
+	option(size(Size), Options), !,
 	content_length(Lines, Size).
 return_size(_, _).
 
 return_fields([], _).
 return_fields([header(Name, Value)|T], Lines) :- !,
-	atom_codes(Name, Codes),
 	(   member(Line, Lines),
-	    phrase(atom_field(Codes, Value), Line)
+	    phrase(atom_field(Name, Value), Line)
 	->  true
 	;   Value = ''
 	),
@@ -545,9 +553,9 @@ return_fields([_|T], Lines) :-
 %	URL after redirections.
 
 return_final_url(Options) :-
-	memberchk(final_url(URL), Options),
+	option(final_url(URL), Options),
 	var(URL), !,
-	memberchk(visited(Parts), Options),
+	option(visited(Parts), Options),
 	parts_uri(Parts, URL).
 return_final_url(_).
 
@@ -577,7 +585,7 @@ transfer_encoding(Lines, Encoding) :-
 	Encoding = Encoding0.
 
 transfer_encoding(Encoding) -->
-	field("transfer-encoding"),
+	field('transfer-encoding'),
 	rest(Encoding).
 
 %%	read_header(+In:stream, -Code:int, -Comment:atom, -Lines:list) is det.
@@ -602,7 +610,7 @@ read_header(In, Code, Comment, Lines) :-
 	).
 read_header(_, 500, 'Invalid reply header', []).
 
-rest_header("", _, []) :- !.		% blank line: end of header
+rest_header([], _, []) :- !.		% blank line: end of header
 rest_header(L0, In, [L0|L]) :-
 	read_line_to_codes(In, L1),
 	rest_header(L1, In, L).
@@ -618,7 +626,7 @@ content_length(Lines, Length) :-
 
 location(Lines, RequestURI) :-
 	member(Line, Lines),
-	phrase(atom_field("location", RequestURI), Line), !.
+	phrase(atom_field(location, RequestURI), Line), !.
 
 first_line(Code, Comment) -->
 	"HTTP/", [_], ".", [_],
@@ -632,17 +640,21 @@ atom_field(Name, Value) -->
 	rest(Value).
 
 content_length(Len) -->
-	field("content-length"),
+	field('content-length'),
 	integer(Len).
 
-field([]) -->
+field(Name) -->
+	{ atom_codes(Name, Codes) },
+	field_codes(Codes).
+
+field_codes([]) -->
 	":",
 	skip_blanks.
-field([H|T]) -->
+field_codes([H|T]) -->
 	[C],
 	{ match_header_char(H, C)
 	},
-	field(T).
+	field_codes(T).
 
 match_header_char(C, C) :- !.
 match_header_char(C, U) :-
@@ -768,7 +780,7 @@ add_authorization(_, Options, Options).
 parse_url_ex(URL, [uri(URL)|Parts]) :-
 	uri_components(URL, Components),
 	phrase(components(Components), Parts),
-	(   memberchk(host(_), Parts)
+	(   option(host(_), Parts)
 	->  true
 	;   domain_error(url, URL)
 	).
@@ -843,7 +855,7 @@ parts_authority(Parts, Auth) :-
 				 uri_authority(User, Password, Host, Port)).
 
 parts_request_uri(Parts, RequestURI) :-
-	memberchk(request_uri(RequestURI), Parts), !.
+	option(request_uri(RequestURI), Parts), !.
 parts_request_uri(Parts, RequestURI) :-
 	url_part(path(Path), Parts, /),
 	ignore(parts_search(Parts, Search)),
@@ -852,14 +864,14 @@ parts_request_uri(Parts, RequestURI) :-
 	uri_components(RequestURI, Data).
 
 parts_search(Parts, Search) :-
-	memberchk(query_string(Search), Parts), !.
+	option(query_string(Search), Parts), !.
 parts_search(Parts, Search) :-
-	memberchk(search(Fields), Parts), !,
+	option(search(Fields), Parts), !,
 	uri_query_components(Search, Fields).
 
 
 parts_uri(Parts, URI) :-
-	memberchk(uri(URI), Parts), !.
+	option(uri(URI), Parts), !.
 parts_uri(Parts, URI) :-
 	parts_scheme(Parts, Scheme),
 	ignore(parts_authority(Parts, Auth)),
@@ -872,13 +884,13 @@ parts_uri(Parts, URI) :-
 url_part(Part, Parts) :-
 	Part =.. [Name,Value],
 	Gen =.. [Name,RawValue],
-	memberchk(Gen, Parts), !,
+	option(Gen, Parts), !,
 	Value = RawValue.
 
 url_part(Part, Parts, Default) :-
 	Part =.. [Name,Value],
 	Gen =.. [Name,RawValue],
-	(   memberchk(Gen, Parts)
+	(   option(Gen, Parts)
 	->  Value = RawValue
 	;   Value = Default
 	).
@@ -896,7 +908,7 @@ update_cookies(_, _, _) :-
 	predicate_property(http:update_cookies(_,_,_), number_of_clauses(0)), !.
 update_cookies(Lines, Parts, Options) :-
 	(   member(Line, Lines),
-	    phrase(atom_field("set_cookie", CookieData), Line),
+	    phrase(atom_field('set_cookie', CookieData), Line),
 	    http:update_cookies(CookieData, Parts, Options),
 	    fail
 	;   true
@@ -918,7 +930,7 @@ update_cookies(Lines, Parts, Options) :-
 %		http:open_options/2.
 %
 %	    http:open_options(Parts, Options) :-
-%		memberchk(host(Host), Parts),
+%		option(host(Host), Parts),
 %		Host \== localhost,
 %		Options = [proxy('proxy.local', 3128)].
 %	    ==
@@ -929,10 +941,18 @@ update_cookies(Lines, Parts, Options) :-
 %	open stream to the HTTP server, Parts is the broken-down request
 %	(see uri_components/2) and Options is the list of options passed
 %	to http_open.  The predicate is called as if using ignore/1.
+%
+%	@see complements http:update_cookies/3.
+%	@see library(http/http_cookies) implements cookie handling on
+%	top of these hooks.
 
-%%	update_cookies(+CookieData, +Parts, +Options) is semidet.
+%%	http:update_cookies(+CookieData, +Parts, +Options) is semidet.
 %
 %	Update the cookie database.  CookieData  is   the  value  of the
 %	=|Set-Cookie|= field, Parts is  the   broken-down  request  (see
 %	uri_components/2) and Options is the list   of options passed to
 %	http_open.
+%
+%	@see complements http:write_cookies
+%	@see library(http/http_cookies) implements cookie handling on
+%	top of these hooks.
