@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -51,6 +51,20 @@
 :- use_module(http_stream).
 
 
+:- predicate_options(http_server/2, 2,
+		     [ port(integer),
+		       tcp_socket(any),
+		       workers(positive_integer),
+		       timeout(number),
+		       keep_alive_timeout(number),
+		       pass_to(system:thread_create/3, 3)
+		     ]).
+:- predicate_options(http_spawn/2, 2,
+		     [ pool(atom),
+		       pass_to(system:thread_create/3, 3),
+		       pass_to(thread_pool:thread_create_in_pool/4, 4)
+		     ]).
+
 /** <module> Threaded HTTP server
 
 This library provides a multi-threaded Prolog-based HTTP server based on
@@ -70,7 +84,6 @@ for details.
 	http_spawn(0, +).
 
 :- dynamic
-	port_option/2,			% Port, Option
 	current_server/5,		% Port, Goal, Thread, Queue, StartTime
 	queue_worker/2,			% Queue, ThreadID
 	queue_options/2.		% Queue, Options
@@ -87,8 +100,9 @@ for details.
 %	Create a server at Port that calls Goal for each parsed request.
 %	Options provide a list of options. Defined options are
 %
-%	| port(?Port)	     | - 	| Port to listen to		      |
-%	| workers(N)	     | 5 	| Define the number of worker threads |
+%	| port(?Port)	     | -	| Port to listen to		      |
+%	| tcp_socket(+Socket)| -	| If provided, use this socket	      |
+%	| workers(N)	     | 5	| Define the number of worker threads |
 %	| timeout(S)	     | 60	| Max inactivity for reading request  |
 %	| keep_alive_timeout | 2	| Drop Keep-Alive connection timeout  |
 %	| local(KBytes)	     | <CommandLine> |				      |
@@ -99,7 +113,6 @@ http_server(Goal, Options) :-
 	strip_module(Goal, Module, G),
 	select_option(port(Port), Options, Options1), !,
 	make_socket(Port, Options1, Options2),
-	set_port_options(Port, Options2),
 	create_workers(Options2),
 	create_server(Module:G, Port, Options2).
 http_server(_Goal, _Options) :-
@@ -113,6 +126,12 @@ http_server(_Goal, _Options) :-
 
 make_socket(Port, Options0, Options) :-
 	make_socket_hook(Port, Options0, Options), !.
+make_socket(Port, Options0, Options) :-
+	option(tcp_socket(_), Options0), !,
+	make_addr_atom('httpd@', Port, Queue),
+	Options = [ queue(Queue)
+		  | Options0
+		  ].
 make_socket(Port, Options0, Options) :-
 	tcp_socket(Socket),
 	tcp_setopt(Socket, reuseaddr),
@@ -156,24 +175,6 @@ create_server(Goal, Port, Options) :-
 	assert(current_server(Port, Goal, Alias, Queue, StartTime)).
 
 
-%%	set_port_options(+Port, +Options) is det.
-%
-%	Register Options for the HTTP server at Port.
-
-set_port_options(Port, Options) :-
-	retractall(port_option(Port, _)),
-	assert_port_options(Options, Port).
-
-assert_port_options([], _).
-assert_port_options([Name=Value|T], Port) :- !,
-	Opt =.. [Name,Value],
-	assert(port_option(Port, Opt)),
-	assert_port_options(T, Port).
-assert_port_options([Opt|T], Port) :- !,
-	assert(port_option(Port, Opt)),
-	assert_port_options(T, Port).
-
-
 %%	http_current_server(:Goal, ?Port) is nondet.
 %
 %	True if Goal is the goal of a server at Port.
@@ -262,7 +263,7 @@ accept_server2(Goal, Options) :-
 	      ;   print_message(error, E),
 		  fail
 	      )
-	  ;   print_message(error, goal_failed(tcp_accept(Socket, Client, Peer)))
+	  ;   print_message(error, goal_failed(tcp_accept(Socket, _, _)))
 	  ).
 
 accept_rethrow_error(http_stop).
@@ -405,14 +406,19 @@ open_client(requeue(In, Out, Goal, ClOpts),
 	memberchk(peer(Peer), ClOpts),
 	option(keep_alive_timeout(KeepAliveTMO), Opts, 2),
 	check_keep_alife_connection(In, KeepAliveTMO, Peer, In, Out).
-open_client(Message, Queue, Goal, In, Out, _Opts,
-	    [ pool(client(Queue, Goal, In, Out))
+open_client(Message, Queue, Goal, In, Out, Opts,
+	    [ pool(client(Queue, Goal, In, Out)),
+	      timeout(Timeout)
 	    | Options
 	    ]) :-
 	catch(open_client(Message, Goal, In, Out, Options),
 	      E, report_error(E)),
-	memberchk(peer(Peer), Options),
-	debug(http(connection), 'Opened connection from ~p', [Peer]).
+	option(timeout(Timeout), Opts, 60),
+	(   debugging(http(connection))
+	->  memberchk(peer(Peer), Options),
+	    debug(http(connection), 'Opened connection from ~p', [Peer])
+	;   true
+	).
 
 
 open_client(Message, Goal, In, Out, Options) :-
@@ -454,15 +460,37 @@ check_keep_alife_connection(In, TMO, Peer, In, Out) :-
 
 %%	done_worker
 %
-%	Called when worker is terminated due to http_workers/2.
+%	Called when worker is terminated  due   to  http_workers/2  or a
+%	(debugging) exception. In  the   latter  case, recreate_worker/2
+%	creates a new worker.
 
 done_worker :-
 	thread_self(Self),
 	thread_property(Self, status(Status)),
-	retract(queue_worker(_Queue, Self)),
-	print_message(informational,
-		      httpd_stopped_worker(Self, Status)).
+	retract(queue_worker(Queue, Self)),
+	(   catch(recreate_worker(Status, Queue), _, fail)
+	->  thread_detach(Self),
+	    print_message(informational,
+			  httpd_restarted_worker(Self))
+	;   print_message(informational,
+			  httpd_stopped_worker(Self, Status))
+	).
 
+%%	recreate_worker(+Status, +Queue) is semidet.
+%
+%	Deal with the possibility that  threads are, during development,
+%	killed with abort/0. We  recreate  the   worker  to  avoid  that
+%	eventually we run out of workers.  If   we  are aborted due to a
+%	halt/0 call, thread_create/3 will raise a permission error.
+
+recreate_worker(exception(Error), Queue) :-
+	recreate_on_error(Error),
+	queue_options(Queue, Options),
+	atom_concat(Queue, '_', AliasBase),
+	create_workers(1, 1, Queue, AliasBase, Options).
+
+recreate_on_error('$aborted').
+recreate_on_error(time_limit_exceeded).
 
 %	thread_httpd:message_level(+Exception, -Level)
 %
@@ -519,12 +547,21 @@ http_process(Goal, In, Out, Options) :-
 	      [Goal, In, Out]),
 	option(timeout(TMO), Options, 60),
 	set_stream(In, timeout(TMO)),
+	set_stream(Out, timeout(TMO)),
 	http_wrapper(Goal, In, Out, Connection,
 		     [ request(Request)
 		     | Options
 		     ]),
 	next(Connection, Request).
 
+next(switch_protocol(SwitchGoal, _SwitchOptions), Request) :- !,
+        memberchk(pool(client(_Queue, _Goal, In, Out)), Request),
+        (   catch(call(SwitchGoal, In, Out), E,
+		  (   print_message(error, E),
+		      fail))
+	->  true
+	;   http_close_connection(Request)
+	).
 next(spawned(ThreadId), _) :- !,
 	debug(http(spawn), 'Handler spawned to thread ~w', [ThreadId]).
 next(Connection, Request) :-
@@ -605,11 +642,13 @@ wrap_spawned(CGI, Goal) :-
 %
 %	Lazy  creation  of  worker-pools  for   the  HTTP  server.  This
 %	predicate calls the hook http:create_pool/1.   If the hook fails
-%	it creates a default pool of size 10.   This should suffice most
-%	typical usecases.
+%	it creates a default pool of size   10. This should suffice most
+%	typical usecases. Note that we  get   a  permission error if the
+%	pool is already created.  We can ignore this.
 
 create_pool(Pool) :-
-	http:create_pool(Pool), !.
+	E = error(permission_error(create, thread_pool, Pool), _),
+	catch(http:create_pool(Pool), E, true).
 create_pool(Pool) :-
 	print_message(informational, httpd(created_pool(Pool))),
 	thread_pool_create(Pool, 10, []).
@@ -625,6 +664,8 @@ create_pool(Pool) :-
 
 prolog:message(httpd_stopped_worker(Self, Status)) -->
 	[ 'Stopped worker ~p: ~p'-[Self, Status] ].
+prolog:message(httpd_restarted_worker(Self)) -->
+	[ 'Replaced aborted worker ~p'-[Self] ].
 prolog:message(httpd(created_pool(Pool))) -->
 	[ 'Created thread-pool ~p of size 10'-[Pool], nl,
 	  'Create this pool at startup-time or define the hook ', nl,
