@@ -1,11 +1,10 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker & Richard O'Keefe
-    E-mail:        wielemaker@science.uva.nl
+    E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2004, University of Amsterdam
+    Copyright (C): 1985-2012, University of Amsterdam
+			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -19,7 +18,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -30,7 +29,7 @@
 */
 
 :- module(sgml_write,
-	  [ html_write/2,       	%          +Data, +Options
+	  [ html_write/2,		%          +Data, +Options
 	    html_write/3,		% +Stream, +Data, +Options
 	    sgml_write/2,		%          +Data, +Options
 	    sgml_write/3,		% +Stream, +Data, +Options
@@ -44,6 +43,22 @@
 :- use_module(library(option)).
 :- use_module(library(error)).
 
+:- predicate_options(xml_write/2, 2, [pass_to(xml_write/3, 3)]).
+:- predicate_options(xml_write/3, 3,
+		     [ dtd(any),
+		       doctype(atom),
+		       public(atom),
+		       system(atom),
+		       header(boolean),
+		       nsmap(list),
+		       indent(nonneg),
+		       layout(boolean),
+		       net(boolean)
+		     ]).
+
+:- multifile
+	xmlns/2.			% NS, URI
+
 /** <module> XML/SGML writer module
 
 This library provides the inverse functionality   of  the sgml.pl parser
@@ -257,7 +272,7 @@ emit_xml_encoding(_, _).
 %	attribute; so the only time this is useful is when it is illegal!
 
 emit_doctype(_Options, Data, Out) :-
-	(   memberchk(element(html,Att,_), Data)
+	(   Data = [_|_], memberchk(element(html,Att,_), Data)
 	;   Data = element(html,Att,_)
 	),
 	memberchk(version=Version, Att),
@@ -274,8 +289,8 @@ emit_doctype(Options, Data, Out) :-
 	    SysId == (-),
 	    \+ memberchk(doctype(_), Options)
 	),
-	(   Data  =   element(DocType,_,_)
-	;   memberchk(element(DocType,_,_), Data)
+	(   Data = element(DocType,_,_)
+	;   Data = [_|_], memberchk(element(DocType,_,_), Data)
 	;   memberchk(doctype(DocType), Options)
 	),
 	!,
@@ -304,7 +319,7 @@ emit([H|T], Out, State) :- !,
 	emit(H, Out, State),
 	emit(T, Out, State).
 emit(CDATA, Out, State) :-
-	atom(CDATA), !,
+	atomic(CDATA), !,
 	sgml_write_content(Out, CDATA, State).
 emit(Element, Out, State) :-
 	\+ \+ emit_element(Element, Out, State).
@@ -312,7 +327,7 @@ emit(Element, Out, State) :-
 emit_element(pi(PI), Out, State) :- !,
 	get_state(State, entity_map, EntityMap),
 	write(Out, <?),
-	write_quoted(Out, PI, "", EntityMap),
+	write_quoted(Out, PI, '', EntityMap),
 	(   get_state(State, dialect, xml) ->
 	    write(Out, ?>)
 	;   write(Out, >)
@@ -459,19 +474,19 @@ content([], Out, Element, State) :- !,	% empty element
 	;   emit_close(Element, Out, State)
 	)
     ).
-content([Atom], Out, Element, State) :-
-	atom(Atom), !,
+content([CDATA], Out, Element, State) :-
+	atomic(CDATA), !,
 	(   get_state(State, dialect, sgml),
 	    get_state(State, net, true),
-	    \+ sub_atom(Atom, _, _, _, /),
-	    atom_length(Atom, Len),
+	    \+ sub_atom(CDATA, _, _, _, /),
+	    write_length(CDATA, Len, []),
 	    Len < 20
 	->  write(Out, /),
-	    sgml_write_content(Out, Atom, State),
+	    sgml_write_content(Out, CDATA, State),
 	    write(Out, /)
 	;/* XML or not NET */
 	    write(Out, >),
-	    sgml_write_content(Out, Atom, State),
+	    sgml_write_content(Out, CDATA, State),
 	    emit_close(Element, Out, State)
 	).
 content(Content, Out, Element, State) :-
@@ -574,16 +589,40 @@ add_missing_ns([H|T], Atts0, Atts) :-
 %
 %	Generate a namespace (NS) identifier for URI.
 
+generate_ns(URI, NS) :-
+	xmlns(NS, URI), !.
 generate_ns(URI, NS) :-
 	default_ns(URI, NS), !.
 generate_ns(_, NS) :-
 	gensym(xns, NS).
 
+%%	xmlns(?NS, ?URI) is nondet.
+%
+%	Hook to define human readable  abbreviations for XML namespaces.
+%	xml_write/3 tries these locations:
+%
+%	  1. This hook
+%	  2. Defaults (see below)
+%	  3. rdf_db:ns/2 for RDF-DB integration
+%
+%	Default XML namespaces are:
+%
+%	  | xsi    | http://www.w3.org/2001/XMLSchema-instance |
+%	  | xs     | http://www.w3.org/2001/XMLSchema          |
+%	  | xhtml  | http://www.w3.org/1999/xhtml	       |
+%	  | soap11 | http://schemas.xmlsoap.org/soap/envelope/ |
+%	  | soap12 | http://www.w3.org/2003/05/soap-envelope   |
+%
+%	@see xml_write/2, rdf_register_ns/2.
+
 :- multifile
 	rdf_db:ns/2.
 
 default_ns('http://www.w3.org/2001/XMLSchema-instance', xsi).
-default_ns('http://www.w3.org/1999/xhtml', xhtml).
+default_ns('http://www.w3.org/2001/XMLSchema',		xs).
+default_ns('http://www.w3.org/1999/xhtml',		xhtml).
+default_ns('http://schemas.xmlsoap.org/soap/envelope/', soap11).
+default_ns('http://www.w3.org/2003/05/soap-envelope',   soap12).
 default_ns(URI, NS) :-
 	rdf_db:ns(NS, URI).
 
@@ -630,13 +669,19 @@ sgml_write_attribute(Out, Values, State) :-
 	is_list(Values), !,
 	get_state(State, entity_map, EntityMap),
 	put_char(Out, '"'),
-	write_quoted_list(Values, Out, """<&>", EntityMap),
+	write_quoted_list(Values, Out, '"<&>', EntityMap),
 	put_char(Out, '"').
 sgml_write_attribute(Out, Value, State) :-
+	is_text(Value), !,
 	get_state(State, entity_map, EntityMap),
 	put_char(Out, '"'),
-	write_quoted(Out, Value, """<&>", EntityMap),
+	write_quoted(Out, Value, '"<&>', EntityMap),
 	put_char(Out, '"').
+sgml_write_attribute(Out, Value, _State) :-
+	number(Value), !,
+	format(Out, '"~w"', [Value]).
+sgml_write_attribute(_, Value, _) :-
+	type_error(sgml_attribute_value, Value).
 
 write_quoted_list([], _, _, _).
 write_quoted_list([H|T], Out, Escape, EntityMap) :-
@@ -649,18 +694,33 @@ write_quoted_list([H|T], Out, Escape, EntityMap) :-
 
 
 sgml_write_content(Out, Value, State) :-
+	is_text(Value), !,
 	get_state(State, entity_map, EntityMap),
-	write_quoted(Out, Value, "<&>", EntityMap).
+	write_quoted(Out, Value, '<&>', EntityMap).
+sgml_write_content(Out, Value, _) :-
+	write(Out, Value).
 
+is_text(Value) :- atom(Value), !.
+is_text(Value) :- string(Value), !.
 
 write_quoted(Out, Atom, Escape, EntityMap) :-
+	atom(Atom), !,
 	atom_codes(Atom, Codes),
 	writeq(Codes, Out, Escape, EntityMap).
+write_quoted(Out, String, Escape, EntityMap) :-
+	string(String), !,
+	string_codes(String, Codes),
+	writeq(Codes, Out, Escape, EntityMap).
+write_quoted(_, String, _, _) :-
+	type_error(atom_or_string, String).
 
 
+%%	writeq(+Text:codes, +Out:stream, +Escape:atom, +Escape:assoc) is det.
+
 writeq([], _, _, _).
 writeq([H|T], Out, Escape, EntityMap) :-
-	(   memberchk(H, Escape)
+	(   char_code(HC, H),
+	    sub_atom(Escape, _, _, _, HC)
 	->  write_entity(H, Out, EntityMap)
 	;   H >= 256
 	->  (   stream_property(Out, encoding(Enc)),
