@@ -1,11 +1,10 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2007, University of Amsterdam
+    Copyright (C): 1985-2013, University of Amsterdam
+			      VU University Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -19,7 +18,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -32,9 +31,10 @@ source should also use format() to produce error messages, etc.
 #include "pl-utf8.h"
 #include <ctype.h>
 
-static char *	formatNumber(bool split, int div, int radix,
-			     bool small, Number n, Buffer out);
-static char *	formatFloat(int how, int arg, Number f, Buffer out);
+static char *	formatInteger(PL_locale *locale, int div, int radix,
+			     bool smll, Number n, Buffer out);
+static char *	formatFloat(PL_locale *locale, int how, int arg,
+			    Number f, Buffer out);
 
 #define MAXRUBBER 100
 
@@ -53,9 +53,9 @@ typedef struct
   struct rubber rub[MAXRUBBER];
 } format_state;
 
-#define BUFSIZE 	1024
-#define DEFAULT 	(-1)
-#define SHIFT   	{ argc--; argv++; }
+#define BUFSIZE		1024
+#define DEFAULT		(-1)
+#define SHIFT		{ argc--; argv++; }
 #define NEED_ARG	{ if ( argc <= 0 ) \
 			  { FMT_ERROR("not enough arguments"); \
 			  } \
@@ -65,6 +65,13 @@ typedef struct
 #define FMT_ARG(c, a)	return (void)Sunlock(fd), \
 			       PL_error(NULL, 0, NULL, \
 					ERR_FORMAT_ARG, c, a)
+#define FMT_EXEPTION()	return (void)Sunlock(fd), FALSE
+
+
+static PL_locale prolog_locale =
+{ 0,0,LOCALE_MAGIC,1,
+  L".", NULL
+};
 
 
 static int
@@ -140,12 +147,6 @@ outstring(format_state *state, const char *s, size_t len)
 }
 
 
-static int
-outstring0(format_state *state, const char *s)
-{ return outstring(state, s, strlen(s));
-}
-
-
 static int
 oututf8(format_state *state, const char *s, size_t len)
 { const char *e = &s[len];
@@ -162,6 +163,12 @@ oututf8(format_state *state, const char *s, size_t len)
 }
 
 
+static int
+oututf80(format_state *state, const char *s)
+{ return oututf8(state, s, strlen(s));
+}
+
+
 static int
 outtext(format_state *state, PL_chars_t *txt)
 { switch(txt->encoding)
@@ -189,7 +196,8 @@ outtext(format_state *state, PL_chars_t *txt)
 #define format_predicates (GD->format.predicates)
 
 static int	update_column(int, Char);
-static bool	do_format(IOSTREAM *fd, PL_chars_t *fmt, int ac, term_t av);
+static bool	do_format(IOSTREAM *fd, PL_chars_t *fmt,
+			  int ac, term_t av, Module m);
 static void	distribute_rubber(struct rubber *, int, int);
 static int	emit_rubber(format_state *state);
 
@@ -201,15 +209,17 @@ static int	emit_rubber(format_state *state);
 word
 pl_format_predicate(term_t chr, term_t descr)
 { int c;
-  Procedure proc;
+  predicate_t proc = NULL;
   Symbol s;
+  int arity;
 
   if ( !PL_get_char_ex(chr, &c, FALSE) )
     fail;
 
   if ( !get_procedure(descr, &proc, 0, GP_CREATE) )
     fail;
-  if ( proc->definition->functor->arity == 0 )
+  PL_predicate_info(proc, NULL, &arity, NULL);
+  if ( arity == 0 )
     return PL_error(NULL, 0, "arity must be > 0", ERR_DOMAIN,
 		    PL_new_atom("format_predicate"),
 		    descr);
@@ -255,8 +265,7 @@ pl_current_format_predicate(term_t chr, term_t descr, control_t h)
   }
   while( (s=advanceTableEnum(e)) )
   { if ( PL_unify_integer(chr, (intptr_t)s->name) &&
-	 unify_definition(contextModule(LD->environment),
-			  descr, ((Procedure)s->value)->definition, 0, 0) )
+	 PL_unify_predicate(descr, (predicate_t)s->value, 0) )
     { PL_close_foreign_frame(fid);
       ForeignRedoPtr(e);
     }
@@ -271,7 +280,7 @@ pl_current_format_predicate(term_t chr, term_t descr, control_t h)
 
 
 static word
-format_impl(IOSTREAM *out, term_t format, term_t Args)
+format_impl(IOSTREAM *out, term_t format, term_t Args, Module m)
 { GET_LD
   term_t argv;
   int argc = 0;
@@ -306,7 +315,7 @@ format_impl(IOSTREAM *out, term_t format, term_t Args)
       break;
   }
 
-  rval = do_format(out, &fmt, argc, argv);
+  rval = do_format(out, &fmt, argc, argv, m);
   PL_free_text(&fmt);
   if ( !endCritical )
     return FALSE;
@@ -317,11 +326,17 @@ format_impl(IOSTREAM *out, term_t format, term_t Args)
 
 word
 pl_format3(term_t out, term_t format, term_t args)
-{ redir_context ctx;
+{ GET_LD
+  redir_context ctx;
   word rc;
+  Module m = NULL;
+  term_t list = PL_new_term_ref();
+
+  if ( !PL_strip_module(args, &m, list) )
+    return FALSE;
 
   if ( (rc=setupOutputRedirect(out, &ctx, FALSE)) )
-  { if ( (rc = format_impl(ctx.stream, format, args)) )
+  { if ( (rc = format_impl(ctx.stream, format, list, m)) )
       rc = closeOutputRedirect(&ctx);
     else
       discardOutputRedirect(&ctx);
@@ -356,7 +371,7 @@ get_chr_from_text(const PL_chars_t *t, int index)
 		********************************/
 
 static bool
-do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
+do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 { GET_LD
   format_state state;			/* complete state */
   int tab_stop = 0;			/* padded tab stop */
@@ -382,6 +397,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
     switch(c)
     { case '~':
 	{ int arg = DEFAULT;		/* Numeric argument */
+	  int mod_colon = FALSE;	/* Used colon modifier */
 					/* Get the numeric argument */
 	  c = get_chr_from_text(fmt, ++here);
 
@@ -416,23 +432,31 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
 	    c = get_chr_from_text(fmt, ++here);
 	  }
 
+	  if ( c == ':' )
+	  { mod_colon = TRUE;
+	    c = get_chr_from_text(fmt, ++here);
+	  }
+
 					/* Check for user defined format */
 	  if ( format_predicates &&
 	       (s = lookupHTable(format_predicates, (void*)((intptr_t)c))) )
-	  { Procedure proc = (Procedure) s->value;
-	    FunctorDef fdef = proc->definition->functor;
-	    term_t av = PL_new_term_refs(fdef->arity);
+	  { predicate_t proc = (predicate_t) s->value;
+	    int arity;
+	    term_t av;
 	    char buf[BUFSIZE];
 	    char *str = buf;
 	    size_t bufsize = BUFSIZE;
-	    unsigned int i;
+	    int i;
+
+	    PL_predicate_info(proc, NULL, &arity, NULL);
+	    av = PL_new_term_refs(arity);
 
 	    if ( arg == DEFAULT )
 	      PL_put_atom(av+0, ATOM_default);
 	    else
 	      PL_put_integer(av+0, arg);
 
-	    for(i=1; i<fdef->arity; i++)
+	    for(i=1; i < arity; i++)
 	    { NEED_ARG;
 	      PL_put_term(av+i, argv);
 	      SHIFT;
@@ -460,7 +484,9 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
 		  if ( !PL_get_text(argv, &txt, CVT_ATOMIC) )
 		    FMT_ARG("a", argv);
 		  SHIFT;
-		  outtext(&state, &txt);
+		  rc = outtext(&state, &txt);
+                  if ( !rc )
+		    goto out;
 		  here++;
 		  break;
 		}
@@ -473,7 +499,9 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
 
 		    SHIFT;
 		    while(times-- > 0)
-		    { outchr(&state, chr);
+		    { rc = outchr(&state, chr);
+		      if ( !rc )
+		        goto out;
 		    }
 		  } else
 		    FMT_ARG("c", argv);
@@ -486,7 +514,11 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
 	      case 'g':			/* shortest of 'f' and 'e' */
 	      case 'G':			/* shortest of 'f' and 'E' */
 		{ number n;
+		  union {
 		  tmp_buffer b;
+		    buffer b1;
+		  } u;
+		  PL_locale *l;
 
 		  NEED_ARG;
 		  if ( !valueExpression(argv, &n PASS_LD) )
@@ -498,11 +530,19 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
 		  }
 		  SHIFT;
 
-		  initBuffer(&b);
-		  formatFloat(c, arg, &n, (Buffer)&b);
+		  if ( c == 'f' && mod_colon )
+		    l = fd->locale;
+		  else
+		    l = &prolog_locale;
+
+		  initBuffer(&u.b);
+		  rc = formatFloat(l, c, arg, &n, &u.b1) != NULL;
 		  clearNumber(&n);
-		  outstring0(&state, baseBuffer(&b, char));
-		  discardBuffer(&b);
+		  if ( rc )
+		    rc = oututf80(&state, baseBuffer(&u.b, char));
+		  discardBuffer(&u.b);
+                  if ( !rc )
+		    goto out;
 		  here++;
 		  break;
 		}
@@ -510,6 +550,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
 	      case 'D':			/* grouped integer */
 	      case 'r':			/* radix number */
 	      case 'R':			/* Radix number */
+	      case 'I':			/* Prolog 1_000_000 */
 		{ number i;
 		  tmp_buffer b;
 
@@ -523,13 +564,42 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
 		    FMT_ARG(f, argv);
 		  }
 		  SHIFT;
-		  if ( arg == DEFAULT )
-		    arg = 0;
 		  initBuffer(&b);
 		  if ( c == 'd' || c == 'D' )
-		  { formatNumber(c == 'D', arg, 10, TRUE, &i, (Buffer)&b);
-		  } else
-		  { if ( arg < 1 || arg > 36 )
+		  { PL_locale ltmp;
+		    PL_locale *l;
+		    static char grouping[] = {3,0};
+
+		    if ( c == 'D' )
+		    { ltmp.thousands_sep = L",";
+		      ltmp.decimal_point = L".";
+		      ltmp.grouping = grouping;
+		      l = &ltmp;
+		    } else if ( mod_colon )
+		    { l = fd->locale;
+		    } else
+		    { l = NULL;
+		    }
+
+		    if ( arg == DEFAULT )
+		      arg = 0;
+		    if ( !formatInteger(l, arg, 10, TRUE, &i, (Buffer)&b) )
+		      FMT_EXEPTION();
+		  } else if ( c == 'I' )
+		  { PL_locale ltmp;
+		    char grouping[2];
+
+		    grouping[0] = (arg == DEFAULT ? 3 : arg);
+		    grouping[1] = '\0';
+		    ltmp.thousands_sep = L"_";
+		    ltmp.grouping = grouping;
+
+		    if ( !formatInteger(&ltmp, 0, 10, TRUE, &i, (Buffer)&b) )
+		      FMT_EXEPTION();
+		  } else			/* r,R */
+		  { if ( arg == DEFAULT )
+		      FMT_ERROR("r,R requires radix specifier");
+		    if ( arg < 1 || arg > 36 )
 		    { term_t r = PL_new_term_ref();
 
 		      PL_put_integer(r, arg);
@@ -537,11 +607,14 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
 		      return PL_error(NULL, 0, NULL, ERR_DOMAIN,
 				      ATOM_radix, r);
 		    }
-		    formatNumber(FALSE, 0, arg, c == 'r', &i, (Buffer)&b);
+		    if ( !formatInteger(NULL, 0, arg, c == 'r', &i, (Buffer)&b) )
+		      FMT_EXEPTION();
 		  }
 		  clearNumber(&i);
-		  outstring0(&state, baseBuffer(&b, char));
+		  rc = oututf80(&state, baseBuffer(&b, char));
 		  discardBuffer(&b);
+		  if ( !rc )
+		    goto out;
 		  here++;
 		  break;
 		}
@@ -552,8 +625,10 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
 		  if ( !PL_get_text(argv, &txt, CVT_LIST|CVT_STRING) &&
 		       !PL_get_text(argv, &txt, CVT_ATOM) ) /* SICStus compat */
 		    FMT_ARG("s", argv);
-		  outtext(&state, &txt);
+		  rc = outtext(&state, &txt);
 		  SHIFT;
+		  if ( !rc )
+		    goto out;
 		  here++;
 		  break;
 		}
@@ -586,8 +661,10 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
 
 		    str = buf;
 		    tellString(&str, &bufsize, ENC_UTF8);
-		    (*f)(argv);
+		    rc = (*f)(argv);
 		    toldString();
+		    if ( !rc )
+		      goto out;
 		    oututf8(&state, str, bufsize);
 		    if ( str != buf )
 		      free(str);
@@ -608,8 +685,10 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
 
 		      str = buf;
 		      tellString(&str, &bufsize, ENC_UTF8);
-		      (*f)(argv);
+		      rc = (*f)(argv);
 		      toldString();
+		      if ( !rc )
+		        goto out;
 		      oututf8(&state, str, bufsize);
 		      if ( str != buf )
 			free(str);
@@ -680,7 +759,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
 		  { FMT_ERROR("not enough arguments");
 		  }
 		  tellString(&str, &bufsize, ENC_UTF8);
-		  rval = callProlog(NULL, argv, PL_Q_CATCH_EXCEPTION, &ex);
+		  rval = callProlog(m, argv, PL_Q_CATCH_EXCEPTION, &ex);
 		  toldString();
 		  oututf8(&state, str, bufsize);
 		  if ( str != buf )
@@ -700,7 +779,9 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
 		  break;
 	        }
 	      case '~':			/* ~ */
-		{ outchr(&state, '~');
+		{ rc = outchr(&state, '~');
+		  if ( !rc )
+		    goto out;
 		  here++;
 		  break;
 		}
@@ -711,7 +792,10 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
 		  if ( c == 'N' && state.column == 0 )
 		    arg--;
 		  while( arg-- > 0 )
-		    outchr(&state, '\n');
+                  { rc = outchr(&state, '\n');
+		    if ( !rc )
+		      goto out;
+                  }
 		  here++;
 		  break;
 		}
@@ -766,7 +850,9 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv)
 	  break;			/* the '~' switch */
 	}
       default:
-	{ outchr(&state, c);
+	{ rc = outchr(&state, c);
+	  if ( !rc )
+	    goto out;
 	  here++;
 	  break;
 	}
@@ -852,107 +938,172 @@ emit_rubber(format_state *state)
 
  ** Fri Aug 19 22:26:41 1988  jan@swivax.UUCP (Jan Wielemaker)  */
 
+static void
+lappend(const wchar_t *l, int def, Buffer out)
+{ if ( l )
+  { const wchar_t *e = l+wcslen(l);
+
+    while (--e >= l)
+    { int c = *e;
+
+      if ( c < 128 )
+      { addBuffer(out, c, char);
+      } else
+      { char buf[6];
+	char *e8, *s;
+
+	e8=utf8_put_char(buf, c);
+	for(s=e8; --s>=buf; )		/* must be reversed as we reverse */
+	{ addBuffer(out, *s, char);	/* in the end */
+	}
+      }
+    }
+  } else
+  { addBuffer(out, def, char);
+  }
+}
+
+static void
+revert_string(char *s, size_t len)
+{ char *e = &s[len-1];
+
+  for(; e>s; s++,e--)
+  { int c = *e;
+
+    *e = *s;
+    *s = c;
+  }
+}
+
 static char *
-formatNumber(bool split, int div, int radix, bool small, Number i,
+formatInteger(PL_locale *locale, int div, int radix, bool smll, Number i,
 	     Buffer out)
-{ switch(i->type)
+{ const char *grouping = NULL;
+
+  if ( !locale )
+  { locale = &prolog_locale;
+  } else
+  { if ( locale->grouping && locale->grouping[0] &&
+	 locale->thousands_sep && locale->thousands_sep[0] )
+      grouping = locale->grouping;
+  }
+
+  switch(i->type)
   { case V_INTEGER:
     { int64_t n = i->value.i;
-      char buf[100];
-      char *tmp, *end, *s;
-      int before = (div == 0);
-      int digits = 0;
-      bool negative = FALSE;
 
-      if ( div+3 > (int)sizeof(buf) )	/* 0.000NNNN with div digits after 0. */
-      { tmp = PL_malloc(div+3);
-	end = tmp+div+3;
-      } else
-      { tmp = buf;
-	end = tmp+sizeof(buf);
-      }
-
-      s = end;				/* i.e. start at the end */
-      *--s = EOS;
-      if ( n < 0 )
-      { n = -n;
-	negative = TRUE;
-      }
       if ( n == 0 && div == 0 )
-      { *--s = '0';
+      { addBuffer(out, '0', char);
       } else
-      { while( n > 0 || div >= 0 )
+      { int before = FALSE;			/* before decimal point */
+	int negative = FALSE;
+	int gsize = 0;
+	int dweight;
+
+	negative = (n < 0);
+
+	while( n != 0 || div >= 0 )
 	{ if ( div-- == 0 && !before )
-	  { *--s = '.';
-	    before = 1;
+	  { if ( !isEmptyBuffer(out) )
+	      lappend(locale->decimal_point, '.', out);
+	    before = TRUE;
+	    if ( grouping )
+	      gsize = grouping[0];
 	  }
-	  if ( split && before && (digits++ % 3) == 0 && digits != 1 )
-	    *--s = ',';
-	  *--s = digitName((int)(n % radix), small);
+
+	  if ( !negative )
+	    dweight = (int)(n % radix);
+	  else
+	    dweight = -(int)(n % -radix);
+
+	  addBuffer(out, digitName(dweight, smll), char);
 	  n /= radix;
+
+	  if ( --gsize == 0 && n != 0 )
+	  { lappend(locale->thousands_sep, ',', out);
+	    if ( grouping[1] == 0 )
+	      gsize = grouping[0];
+	    else if ( grouping[1] == CHAR_MAX )
+	      gsize = 0;
+	    else
+	      gsize = *++grouping;
+	  }
 	}
 	if ( negative )
-	  *--s = '-';
+	  addBuffer(out, '-', char);
       }
 
-      addMultipleBuffer(out, s, end-s, char);
-      if ( tmp != buf )
-	PL_free(tmp);
+      revert_string(baseBuffer(out, char), entriesBuffer(out, char));
+      addBuffer(out, EOS, char);
 
       return baseBuffer(out, char);
     }
 #ifdef O_GMP
     case V_MPZ:
-    { size_t len = mpz_sizeinbase(i->value.mpz, radix);
+    { GET_LD
+      size_t len = mpz_sizeinbase(i->value.mpz, radix);
       char tmp[256];
       char *buf;
+      int rc = TRUE;
 
       if ( len+2 > sizeof(tmp) )
 	buf = PL_malloc(len+2);
       else
 	buf = tmp;
 
-      mpz_get_str(buf, radix, i->value.mpz);
-      if ( !small && radix > 10 )
+      EXCEPTION_GUARDED({ LD->gmp.persistent++;
+			  mpz_get_str(buf, radix, i->value.mpz);
+			  LD->gmp.persistent--;
+			},
+			{ LD->gmp.persistent--;
+			  rc = PL_rethrow();
+			});
+      if ( !rc )
+	return NULL;
+
+      if ( !smll && radix > 10 )
       { char *s;
 
 	for(s=buf; *s; s++)
 	  *s = toupper(*s);
       }
-      if ( split || div > 0 )
-      { int before = (int)(len-div);
-	int leading;
-	char *s = buf;
 
-	if ( *s == '-' )
-	{ addBuffer(out, *s, char);
-	  s++;
-	}
-	if ( split )
-	{ leading = before % 3;
-	  if ( leading == 0 )
-	    leading = 3;
-	} else
-	{ leading = (int)len;
-	}
-	for(; *s; s++)
-	{ if ( before-- == 0 && div > 0 )
-	  { addBuffer(out, '.', char);
-	  } else if ( leading-- == 0 && before > 0 )
-	  { addBuffer(out, ',', char);
-	    leading = 2;
+      if ( grouping || div > 0 )
+      { int before = FALSE;			/* before decimal point */
+	int gsize = 0;
+	char *e = buf+strlen(buf)-1;
+
+	while(e >= buf || div >= 0)
+	{ if ( div-- == 0 && !before )
+	  { if ( !isEmptyBuffer(out) )
+	      lappend(locale->decimal_point, '.', out);
+	    before = TRUE;
+	    if ( grouping )
+	      gsize = grouping[0];
+	  }
+
+	  addBuffer(out, *e, char);
+	  e--;
+
+	  if ( --gsize == 0 && e >= buf && *e != '-' )
+	  { lappend(locale->thousands_sep, ',', out);
+	    if ( grouping[1] == 0 )
+	      gsize = grouping[0];
+	    else if ( grouping[1] == CHAR_MAX )
+	      gsize = 0;
+	    else
+	      gsize = *++grouping;
 	  }
-	  addBuffer(out, *s, char);
 	}
-	addBuffer(out, EOS, char);
+	revert_string(baseBuffer(out, char), entriesBuffer(out, char));
       } else
       { addMultipleBuffer(out, buf, strlen(buf), char);
-	addBuffer(out, EOS, char);
       }
 
       if ( buf != tmp )
 	PL_free(buf);
 
+      addBuffer(out, EOS, char);
       return baseBuffer(out, char);
     }
 #endif /*O_GMP*/
@@ -963,8 +1114,187 @@ formatNumber(bool split, int div, int radix, bool small, Number i,
 }
 
 
+static int
+countGroups(const char *grouping, int len)
+{ int groups = 0;
+  int gsize = grouping[0];
+
+  while(len>0)
+  { len -= gsize;
+
+    if ( len > 0 )
+      groups++;
+
+    if ( grouping[1] == 0 )
+    { if ( len > 1 )
+	groups += (len-1)/grouping[0];
+      return groups;
+    } else if ( grouping[1] == CHAR_MAX )
+    { return groups;
+    } else
+    { gsize = *++grouping;
+    }
+  }
+
+  return groups;
+}
+
+
+static int
+ths_to_utf8(char *u8, const wchar_t *s, size_t len)
+{ char *e = u8+len-7;
+
+  for( ; u8<e && *s; s++)
+    u8 = utf8_put_char(u8,*s);
+  *u8 = EOS;
+
+  return *s == 0;
+}
+
+
+static int
+same_decimal_point(PL_locale *l1, PL_locale *l2)
+{ if ( l1->decimal_point && l2->decimal_point &&
+       wcscmp(l1->decimal_point, l2->decimal_point) == 0 )
+    return TRUE;
+  if ( !l1->decimal_point && !l2->decimal_point )
+    return TRUE;
+
+  return FALSE;
+}
+
+
+static int
+utf8_dp(PL_locale *l, char *s, int *len)
+{ if ( l->decimal_point )
+  { if ( !ths_to_utf8(s, l->decimal_point, 20) )
+      return FALSE;
+    *len = strlen(s);
+  } else
+  { *s++ = '.';
+    *s = EOS;
+    *len = 1;
+  }
+
+  return TRUE;
+}
+
+
+/* localizeDecimalPoint() replaces the decimal point as entered by the
+   local sensitive print functions by the one in the specified locale.
+   This is overly complicated. Needs more testing, in particular for
+   locales with (in UTF-8) multibyte decimal points.
+*/
+
+static int
+localizeDecimalPoint(PL_locale *locale, Buffer b)
+{ if ( locale == GD->locale.default_locale ||
+       same_decimal_point(GD->locale.default_locale, locale) )
+    return TRUE;
+
+  if ( locale->decimal_point && locale->decimal_point[0] )
+  { char *s = baseBuffer(b, char);
+    char *e;
+    char dp[20];  int dplen;
+    char ddp[20]; int ddplen;
+
+    if ( !utf8_dp(locale, dp, &dplen) ||
+	 !utf8_dp(GD->locale.default_locale, ddp, &ddplen) )
+      return FALSE;
+
+    if ( *s == '-' )
+      s++;
+    for(e=s; *e && isDigit(*e); e++)
+      ;
+
+    if ( strncmp(e, ddp, ddplen) == 0 )
+    { if ( dplen == ddplen )
+      { memcpy(e, dp, dplen);
+      } else
+      { char *ob = baseBuffer(b, char);
+	if ( dplen > ddplen && !growBuffer(b, dplen-ddplen) )
+	  return PL_no_memory();
+	e += baseBuffer(b, char) - ob;
+
+	memmove(&e[dplen-ddplen], e, strlen(e)+1);
+	memcpy(e, dp, dplen);
+      }
+    }
+  }
+
+  return TRUE;
+}
+
+
+static int
+groupDigits(PL_locale *locale, Buffer b)
+{ if ( locale->thousands_sep && locale->thousands_sep[0] &&
+       locale->grouping && locale->grouping[0] )
+  { char *s = baseBuffer(b, char);
+    char *e;
+    int groups;
+
+    if ( *s == '-' )
+      s++;
+    for(e=s; *e && isDigit(*e); e++)
+      ;
+
+    groups = countGroups(locale->grouping, (int)(e-s));
+
+    if ( groups > 0 )
+    { char *o;
+      char *grouping = locale->grouping;
+      int gsize = grouping[0];
+      char ths[20];
+      int thslen;
+
+      if ( !ths_to_utf8(ths, locale->thousands_sep, sizeof(ths)) )
+	return FALSE;
+      thslen = strlen(ths);
+
+      if ( !growBuffer(b, thslen*groups) )
+	return PL_no_memory();
+      memmove(&e[groups*thslen], e, strlen(e)+1);
+
+      e--;
+      for(o=e+groups*thslen; e>=s; )
+      { *o-- = *e--;
+	if ( --gsize == 0 && e>=s )
+	{ o -= thslen-1;
+	  memcpy(o, ths, thslen);
+	  o--;
+	  if ( grouping[1] == 0 )
+	    gsize = grouping[0];
+	  else if ( grouping[1] == CHAR_MAX )
+	    gsize = 0;
+	  else
+	    gsize = *++grouping;
+	}
+      }
+    }
+  }
+
+  return TRUE;
+}
+
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+formatFloat(PL_locale *locale, int how, int arg, Number f, Buffer out)
+
+formats a floating point  number  to  a   buffer.  `How'  is  the format
+specifier ([eEfgG]), `arg' the argument.
+
+MPZ/MPQ numbers printed using the format specifier `f' are written using
+the following algorithm, courtesy of Jan Burse:
+
+  Given: A rational n/m
+  Seeked: The ration rounded to d fractional digits.
+  Algorithm: Compute (n*10^d+m//2)//m, and place period at d.
+
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 static char *
-formatFloat(int how, int arg, Number f, Buffer out)
+formatFloat(PL_locale *locale, int how, int arg, Number f, Buffer out)
 { if ( arg == DEFAULT )
     arg = 6;
 
@@ -972,50 +1302,264 @@ formatFloat(int how, int arg, Number f, Buffer out)
   {
 #ifdef O_GMP
     mpf_t mpf;
+    mpz_t t1, t2;
+    int neg;
+    int exp;
+
     case V_MPZ:
-      mpf_init2(mpf, arg*4);
-      mpf_set_z(mpf, f->value.mpz);
-      goto print;
+    { switch(how)
+      { case 'f':
+        { mpz_init(t1);
+          mpz_init(t2);
+          mpz_ui_pow_ui(t1, 10, arg);
+          mpz_mul(t1, f->value.mpz, t1);
+          neg = (mpz_cmp_ui(t1, 0) < 0) ? 1 : 0;
+          mpz_abs(t1, t1);
+          goto print_mpz_f;
+        }
+        case 'e':
+	case 'E':
+        { mpz_init(t1);
+          mpz_init(t2);
+
+          mpz_set(t1, f->value.mpz);
+          neg = (mpz_cmp_ui(t1, 0) < 0) ? 1 : 0;
+          mpz_abs(t1, t1);
+
+          if (mpz_cmp_ui(t1, 0) == 0)
+          { exp = 0;
+          } else
+          { exp = mpz_sizeinbase(t1, 10)-1;  /* guess exponent */
+            mpz_ui_pow_ui(t2, 10, exp);
+            if (mpz_cmp(t2, t1) > 0) exp--;  /* correct exponent */
+
+            if (exp > arg)
+            { mpz_ui_pow_ui(t2, 10, (exp-arg-1));
+              mpz_tdiv_q(t1, t1, t2);
+              mpz_add_ui(t1, t1, 5);
+              mpz_tdiv_q_ui(t1, t1, 10);
+            } else
+            { mpz_ui_pow_ui(t2, 10, arg-exp);
+              mpz_mul(t1, t1, t2);
+            }
+          }
+          mpz_set_ui(t2, exp);
+          goto print_mpz_e;
+        }
+        case 'g':
+	case 'G':
+        { mpf_init2(mpf, arg*4);
+          mpf_set_z(mpf, f->value.mpz);
+          goto print_mpf;
+        }
+      }
+    }
     case V_MPQ:
     { char tmp[12];
       int size;
-      int written;
+      int written = 0;
       int fbits;
+      int digits = 0;
+      int padding = 0;
 
       switch(how)
       { case 'f':
-	case 'g':
+        { mpz_init(t1);
+          mpz_init(t2);
+          mpz_ui_pow_ui(t1, 10, arg);
+          mpz_mul(t1, mpq_numref(f->value.mpq), t1);
+          mpz_tdiv_q_ui(t2, mpq_denref(f->value.mpq), 2);
+          if (mpq_cmp_ui(f->value.mpq, 0, 1) < 0)
+          { mpz_sub(t1, t1, t2);
+            neg=1;
+          } else
+          { mpz_add(t1, t1, t2);
+            neg=0;
+          }
+          mpz_tdiv_q(t1, t1, mpq_denref(f->value.mpq));
+          mpz_abs(t1, t1);
+
+        print_mpz_f:
+
+          if (mpz_cmp_ui(t1, 0) != 0)
+          { size = mpz_sizeinbase(t1, 10) + 1; /* reserve for <null> */
+            if ( !growBuffer(out, size) )
+            { PL_no_memory();
+              return NULL;
+            }
+            digits = written = gmp_snprintf(baseBuffer(out, char), size, "%Zd", t1);
+          }
+          if (digits <= arg) padding = (arg-digits+1);
+
+          size = digits;
+          if (neg) size++;               /* leading - */
+          if (arg) size++;               /* decimal point */
+          if (padding) size += padding;  /* leading '0's */
+          size++;                        /* NULL terminator */
+
+          if ( !growBuffer(out, size) )
+          { PL_no_memory();
+            return NULL;
+          }
+
+          if (!digits)
+          { memset(out->base, '\0', 1);
+          }
+
+          if (padding)
+          { memmove(out->base+padding, out->base, written+1);
+            memset(out->base, '0', padding);
+            written += padding;
+          }
+
+          if (arg)
+          { memmove(out->base+written-(arg-1), out->base+written-arg, arg+1);
+            if ( locale->decimal_point && locale->decimal_point[0] )
+              *(out->base+written-arg) = locale->decimal_point[0];
+            else
+              *(out->base+written-arg) = '.';
+            written++;
+          }
+
+          if (neg)
+          { memmove(out->base+1, out->base, written+1);
+            memset(out->base, '-', 1);
+            written++;
+          }
+
+          out->top = out->base + written;
+          mpz_clear(t1);
+          mpz_clear(t2);
+          break;
+        }
+        case 'e':
+	case 'E':
+        { mpz_init(t1);
+          mpz_init(t2);
+
+          if (mpz_cmpabs(mpq_numref(f->value.mpq), mpq_denref(f->value.mpq)) >= 0)
+          { mpz_tdiv_q(t1, mpq_numref(f->value.mpq), mpq_denref(f->value.mpq));
+            exp = mpz_sizeinbase(t1, 10)-1;     /* guess exponent */
+            mpz_ui_pow_ui(t2, 10, exp);
+            if (mpz_cmpabs(t2, t1) > 0) exp--;  /* correct exponent */
+          } else
+          { mpq_t tq1, tq2;
+            mpq_init(tq1);
+            mpq_init(tq2);
+            mpz_set(t1, mpq_numref(f->value.mpq));
+            mpz_abs(t1, t1);
+            mpz_tdiv_q(t1, mpq_denref(f->value.mpq), t1);
+            exp = 0-mpz_sizeinbase(t1, 10);     /* guess exponent */
+            mpz_ui_pow_ui(t2, 10, abs(exp)-1);
+            mpq_set_z(tq1, t1);
+            mpq_inv(tq2, f->value.mpq);
+            if (mpz_cmpabs(t2, t1) == 0 && mpq_cmp(tq1, tq2)==0) exp++;
+            if (mpz_cmpabs(t2, t1) > 0) exp++;  /* correct exponent */
+            mpq_clear(tq1);
+            mpq_clear(tq2);
+          }
+
+          mpz_ui_pow_ui(t1, 10, abs(arg-exp));
+          if (mpz_cmpabs(mpq_numref(f->value.mpq), mpq_denref(f->value.mpq)) >= 0 && exp > arg)
+          { mpz_tdiv_q(t1, mpq_numref(f->value.mpq), t1);
+          } else
+          { mpz_mul(t1, mpq_numref(f->value.mpq), t1);
+          }
+          mpz_tdiv_q_ui(t2, mpq_denref(f->value.mpq), 2);
+          if (mpq_cmp_ui(f->value.mpq, 0, 1) < 0)
+          { mpz_sub(t1, t1, t2);
+            neg=1;
+          } else
+          { mpz_add(t1, t1, t2);
+            neg=0;
+          }
+          mpz_tdiv_q(t1, t1, mpq_denref(f->value.mpq));
+          mpz_abs(t1, t1);
+
+          mpz_ui_pow_ui(t2, 10, arg+1);
+          if (mpz_cmpabs(t2, t1) == 0)
+          { mpz_tdiv_q_ui(t1, t1, 10);
+            exp++;
+          }
+          mpz_set_ui(t2, abs(exp));
+
+        print_mpz_e:
+
+          if (mpz_cmp_ui(t1, 0) == 0)
+            size = arg+7; /* reserve for 0.+e00<null> */
+          else
+            size = mpz_sizeinbase(t1, 10) + mpz_sizeinbase(t2, 10) + 6; /* reserve for -.e+0<null> */
+
+          if ( !growBuffer(out, size) )
+          { PL_no_memory();
+            return NULL;
+          }
+          written = gmp_snprintf(baseBuffer(out, char), size, "%Zd%c%+03d", t1, how, exp);
+
+          if (mpz_cmp_ui(t1, 0) == 0)
+          { memmove(out->base+arg, out->base, written+1);
+            memset(out->base, '0', arg);
+            written += arg;
+          }
+
+          if (arg)
+          { memmove(out->base+2, out->base+1, written+1);
+            if ( locale->decimal_point && locale->decimal_point[0] )
+              *(out->base+1) = locale->decimal_point[0];
+            else
+              *(out->base+1) = '.';
+            written++;
+          }
+
+          if (neg)
+          { memmove(out->base+1, out->base, written+1);
+            memset(out->base, '-', 1);
+            written++;
+          }
+
+          out->top = out->base + written;
+          mpz_clear(t1);
+          mpz_clear(t2);
+          return baseBuffer(out, char);
+        }
+        case 'g':
 	case 'G':
-	{ mpz_t iv;
+        { switch(how)
+          { case 'g':
+            case 'G':
+            { mpz_t iv;
+              mpz_init(iv);
+              mpz_set_q(iv, f->value.mpq);
+              fbits = (int)mpz_sizeinbase(iv, 2) + 4*arg;
+              mpz_clear(iv);
+              break;
+            }
+            default:
+              fbits = 4*arg;
+          }
+          mpf_init2(mpf, fbits);
+          mpf_set_q(mpf, f->value.mpq);
 
-	  mpz_init(iv);
-	  mpz_set_q(iv, f->value.mpq);
-	  fbits = (int)mpz_sizeinbase(iv, 2) + 4*arg;
-	  mpz_clear(iv);
-	  break;
-	}
-	default:
-	  fbits = 4*arg;
+        print_mpf:
+          Ssprintf(tmp, "%%.%dF%c", arg, how);
+          size = 0;
+          written = arg+4;
+          while(written >= size)
+          { size = written+1;
+
+            if ( !growBuffer(out, size) )	/* reserve for -.e<null> */
+            { PL_no_memory();
+              return NULL;
+            }
+            written = gmp_snprintf(baseBuffer(out, char), size, tmp, mpf);
+          }
+          mpf_clear(mpf);
+          out->top = out->base + written;
+
+          break;
+        }
       }
-
-      mpf_init2(mpf, fbits);
-      mpf_set_q(mpf, f->value.mpq);
-
-    print:
-      Ssprintf(tmp, "%%.%dF%c", arg, how);
-      size = 0;
-      written = arg+4;
-      while(written >= size)
-      { size = written+1;
-
-	if ( !growBuffer(out, size) )	/* reserve for -.e<null> */
-	  outOfCore();
-	written = gmp_snprintf(baseBuffer(out, char), size, tmp, mpf);
-      }
-      mpf_clear(mpf);
-      out->top = out->base + written;
-
-      return baseBuffer(out, char);
+      break;
     }
 #endif
     case V_INTEGER:
@@ -1031,14 +1575,25 @@ formatFloat(int how, int arg, Number f, Buffer out)
       { size = written+1;
 
 	if ( !growBuffer(out, size) )
-	  outOfCore();
+	{ PL_no_memory();
+	  return NULL;
+	}
 	written = snprintf(baseBuffer(out, char), size, tmp, f->value.f);
       }
       out->top = out->base + written;
 
-      return baseBuffer(out, char);
+      break;
     }
+    default:
+      assert(0);
+      return NULL;
   }
 
-  return NULL;
+  if ( locale )
+  { if ( !localizeDecimalPoint(locale, out) ||
+	 !groupDigits(locale, out) )
+      return NULL;
+  }
+
+  return baseBuffer(out, char);
 }
