@@ -107,7 +107,7 @@ PL_get_frame(term_t r, LocalFrame *fr)
 }
 
 
-static void
+void
 PL_put_choice(term_t t, Choice ch)
 { GET_LD
 
@@ -1250,7 +1250,7 @@ traceInterception(LocalFrame frame, Choice bfr, int port, Code PC)
 	{ debugstatus.retryFrame = fr;
 	  rval = ACTION_RETRY;
 	} else
-	  PL_warning("prolog_trace_interception/3: bad argument to retry/1");
+	  PL_warning("prolog_trace_interception/4: bad argument to retry/1");
       }
     }
 
@@ -1377,7 +1377,7 @@ PL_describe_context(pl_context_t *c, char *buf, size_t len)
     if ( c->pc >= fr->predicate->codes &&
 	 c->pc < &fr->predicate->codes[fr->predicate->codes[-1]] )
     { return printed+snprintf(buf, len, "[PC=%ld in supervisor]",
-			      (c->pc - fr->predicate->codes));
+			      (long)(c->pc - fr->predicate->codes));
     }
 
     if ( false(fr->predicate, P_FOREIGN) )
@@ -2118,7 +2118,7 @@ prolog_frame_attribute(term_t frame, term_t what, term_t value)
       if ( tTop+arity > tMax )
       { int rc;
 
-	if ( !(rc=ensureTrailSpace(arity)) != TRUE )
+	if ( (rc=ensureTrailSpace(arity)) != TRUE )
 	  return raiseStackOverflow(rc);
       }
 
